
""" Action for running a test locally or on a remote machine """

import os
import logging
import subprocess
import sys
import signal
import pipes
from texttestlib import plugins
from texttestlib.jobprocess import killProcessAndChildren
from time import sleep
from threading import Lock, Timer
from locale import getpreferredencoding

plugins.addCategory("killed", "killed", "were terminated before completion")


class Running(plugins.TestState):
    defaultBriefText = "RUN"

    def __init__(self, execMachines, freeText="", briefText="", lifecycleChange="start"):
        plugins.TestState.__init__(self, "running", freeText, briefText, started=1,
                                   executionHosts=execMachines, lifecycleChange=lifecycleChange)

    def makeModifiedState(self, newRunStatus, newDetails, lifecycleChange):
        currRunStatus = self.briefText.split()[0]
        if newRunStatus != currRunStatus:
            currFreeTextStatus = self.freeText.splitlines()[0].rsplit(" ", 2)[0]
            newFreeText = self.freeText.replace(currFreeTextStatus, newDetails)
            newBriefText = self.briefText.replace(currRunStatus, newRunStatus)
            return self.__class__(self.executionHosts, newFreeText, newBriefText, lifecycleChange)


class Killed(plugins.TestState):
    def __init__(self, briefText, freeText, prevState):
        plugins.TestState.__init__(self, "killed", briefText=briefText, freeText=freeText,
                                   started=1, completed=1, executionHosts=prevState.executionHosts)
        # Cache running information, it can be useful to have this available...
        self.prevState = prevState
        self.failedPrediction = self


class RunTest(plugins.Action):
    def __init__(self):
        self.diag = logging.getLogger("run test")
        self.killDiag = logging.getLogger("kill processes")
        self.currentProcess = None
        self.currentTimer = None
        self.killedTests = []
        self.killSignal = None
        self.lock = Lock()

    def __repr__(self):
        return "Running"

    def __call__(self, test):
        return self.runTest(test)

    def changeToRunningState(self, test):
        execMachines = test.state.executionHosts
        self.diag.info("Changing " + repr(test) + " to state Running on " + repr(execMachines))
        briefText = self.getBriefText(execMachines)
        freeText = "Running on " + ",".join(execMachines)
        newState = Running(execMachines, briefText=briefText, freeText=freeText)
        test.changeState(newState)

    def getBriefText(self, execMachinesArg):
        # Default to not bothering to print the machine name: all is local anyway
        return ""

    def startTimer(self, timer):
        self.currentTimer = timer
        self.currentTimer.start()

    def runMultiTimer(self, timeout, method, args):
        # Break the timer up into 5 sub-timers
        # The point is to prevent timing out too early if the process gets suspended
        subTimerCount = 5  # whatever
        subTimerTimeout = float(timeout) / subTimerCount
        timer = Timer(subTimerTimeout, method, args)
        for _ in range(subTimerCount - 1):
            timer = Timer(subTimerTimeout, self.startTimer, [timer])
        self.startTimer(timer)

    def runTest(self, test):
        # If previous test (or test run) timed out, don't leave killSignal set, other tests can still run here
        if self.killSignal == "timeout":
            self.killSignal = None
            if test in self.killedTests:
                self.killedTests.remove(test)

        self.describe(test)
        machine = test.app.getRunMachine()
        killTimeout = test.getConfigValue("kill_timeout")
        for postfix in self.getTestRunPostfixes(test):
            if postfix:
                # Checks for support processes like virtual displays, restarts if needed
                test.notify("TestProcessComplete")

            process = self.getTestProcess(test, machine, postfix)
            self.registerProcess(test, process)
            if not postfix:
                # Don't claim to be running until we are, i.e. the process has started
                self.changeToRunningState(test)

            if killTimeout and not test.app.isRecording() and not test.app.isActionReplay():
                self.runMultiTimer(killTimeout, self.kill, (test, "timeout"))
                self.wait(process)
                self.currentTimer.cancel()
                self.currentTimer = None
            else:
                self.wait(process)
            self.checkAndClear(test, postfix)
            if self.killSignal is not None:
                break  # Don't start other processes

    def getTestRunPostfixes(self, test):
        postfixes = [""]
        for postfix in test.getConfigValue("extra_test_process_postfix"):
            if any((test.getFileName(stem + postfix) for stem in test.defFileStems())):
                postfixes.append(postfix)
        return postfixes

    def registerProcess(self, test, process):
        self.lock.acquire()
        self.currentProcess = process
        if test in self.killedTests:
            self.killProcess(test)
        self.lock.release()

    def storeReturnCode(self, test, code, postfix):
        file = open(test.makeTmpFileName("exitcode" + postfix), "w")
        file.write(str(code) + "\n")
        file.close()

    def checkAndClear(self, test, postfix):
        returncode = self.currentProcess.returncode
        self.diag.info("Process terminated with return code " + repr(returncode))
        if os.name == "posix" and test not in self.killedTests and returncode < 0:
            # Process externally killed, but we haven't been notified. Wait for a while to see if we get kill notification
            self.waitForKill()

        self.lock.acquire()
        self.currentProcess = None
        if test in self.killedTests:
            self.changeToKilledState(test)
        elif returncode:  # Don't bother to store return code when tests are killed, it isn't interesting
            self.storeReturnCode(test, returncode, postfix)

        self.lock.release()

    def waitForKill(self):
        for _ in range(10):
            sleep(0.2)
            if self.killSignal is not None:
                return

    def changeToKilledState(self, test):
        self.diag.info("Killing test " + repr(test) + " in state " + test.state.category)
        briefText, fullText = self.getKillInfo(test)
        freeText = "Test " + fullText + "\n"
        test.changeState(Killed(briefText, freeText, test.state))

    def getKillInfo(self, test):
        if self.killSignal is None or self.killSignal == signal.SIGINT or (self.killSignal == 0 and os.name == "nt"):
            return self.getExplicitKillInfo()
        elif self.killSignal == "timeout":
            return "TIMEOUT", "exceeded wallclock time limit of " + str(test.getConfigValue("kill_timeout")) + " seconds"
        elif self.killSignal == signal.SIGTERM:
            return "TERMINATED", "terminated via SIGTERM signal"
        elif hasattr(signal, "SIGXCPU") and self.killSignal == signal.SIGXCPU:
            return "CPULIMIT", "exceeded maximum cpu time allowed"
        else:
            return self.getKillInfoOtherSignal(test)

    def getSignalName(self, sigNum):
        for entry in dir(signal):
            if entry.startswith("SIG") and not entry.startswith("SIG_"):
                number = getattr(signal, entry)
                if number == sigNum:
                    return entry
        return str(sigNum)

    def getKillInfoOtherSignal(self, test):
        briefText = self.getSignalName(self.killSignal)
        return briefText, "terminated by signal " + briefText

    def getExplicitKillInfo(self):
        timeStr = plugins.localtime("%H:%M")
        return "KILLED", "killed explicitly at " + timeStr

    def kill(self, test, sig):
        self.lock.acquire()
        self.killedTests.append(test)
        self.killSignal = sig
        if self.currentProcess is not None:
            self.killProcess(test)
        self.lock.release()

    def killProcess(self, test):
        machine = test.app.getRunMachine()
        if machine != "localhost" and test.getConfigValue("remote_shell_program") == "ssh":
            self.killRemoteProcess(test, machine)
        self.killDiag.info("Killing running test (process id " + str(self.currentProcess.pid) + ")")
        killProcessAndChildren(self.currentProcess.pid, cmd=test.getConfigValue("kill_command"))

    def killRemoteProcess(self, test, machine):
        tmpDir = self.getTmpDirectory(test)
        remoteScript = os.path.join(tmpDir, "kill_test.sh")
        test.app.runCommandOn(machine, ["sh", plugins.quote(remoteScript)])

    def wait(self, process):
        try:
            plugins.retryOnInterrupt(process.wait)
        except OSError:  # pragma: no cover - workaround for Python bugs only
            pass  # safest, as there are python bugs in this area

    def getRunDescription(self, test):
        commandArgs = self.getLocalExecuteCmdArgs(test, makeDirs=False)
        text = "Command Line   : " + plugins.commandLineString(commandArgs) + "\n"
        text += "\nEnvironment variables :\n"
        for var, value in self.getEnvironmentChanges(test):
            text += var + "=" + value + "\n"
        return text

    def getEnvironmentChanges(self, test, postfix=""):
        testEnv = self.getTestRunEnvironment(test, postfix)
        changes = []
        # copy_test_path might be handled in a grid/cloud slave
        # Can't guarantee we get the right value here. So we fake it...
        copyVars = test.getConfigValue("copy_test_path", expandVars=False)
        for var, value in list(testEnv.items()):
            if test.app.hasChanged(var, value):
                if "$" + var in copyVars or "${" + var + "}" in copyVars:
                    value = test.makeTmpFileName(os.path.basename(value), forComparison=False)
                changes.append((var, value))

        return sorted(changes)

    def getTestRunEnvironment(self, test, postfix):
        testEnv = test.getRunEnvironment()
        if postfix and "USECASE_RECORD_SCRIPT" in testEnv:
            # Redirect usecase variables if needed
            self.fixUseCaseVariables(testEnv, postfix)
        return testEnv

    def fixUseCaseVariables(self, testEnv, postfix):
        for varName in ["USECASE_RECORD_SCRIPT", "USECASE_REPLAY_SCRIPT"]:
            if varName in testEnv:
                testEnv[varName] = self.rreplace(testEnv.get(varName), "usecase", "usecase" + postfix)

    def rreplace(self, s, old, new):
        # Swiped from http://stackoverflow.com/questions/2556108/how-to-replace-the-last-occurence-of-an-expression-in-a-string
        parts = s.rsplit(old, 1)
        return new.join(parts)

    def getTestProcess(self, test, machine, postfix=""):
        commandArgs = self.getExecuteCmdArgs(test, machine, postfix)
        self.diag.info("Running test with args : " + repr(commandArgs))
        namingScheme = test.app.getConfigValue("filename_convention_scheme")
        stdoutStem = test.app.getStdoutName(namingScheme) + postfix
        stderrStem = test.app.getStderrName(namingScheme) + postfix
        inputStem = test.app.getStdinName(namingScheme) + postfix
        testEnv = self.getTestRunEnvironment(test, postfix)
        try:
            return subprocess.Popen(commandArgs, preexec_fn=self.getPreExecFunction(),
                                    stdin=open(self.getInputFile(test, inputStem)), cwd=test.getDirectory(temporary=1, local=1),
                                    stdout=self.makeFile(test, stdoutStem), stderr=self.makeFile(test, stderrStem),
                                    env=testEnv, startupinfo=self.getProcessStartUpInfo(test))
        except OSError:
            message = "OS-related error starting the test command - probably cannot find the program " + \
                repr(commandArgs[0])
            raise plugins.TextTestError(message)

    def getProcessStartUpInfo(self, test):
        # Used for hiding the windows if we're on Windows!
        if os.name == "nt" and test.getConfigValue("virtual_display_hide_windows") == "true" and \
                not test.app.isRecording() and not test.app.isActionReplay():
            return plugins.getHideStartUpInfo()

    def getPreExecFunction(self):
        if os.name == "posix":  # pragma: no cover - only run in the subprocess!
            return self.ignoreJobControlSignals

    def ignoreJobControlSignals(self):  # pragma: no cover - only run in the subprocess!
        for signum in [signal.SIGQUIT, signal.SIGUSR1, signal.SIGUSR2, signal.SIGXCPU]:
            signal.signal(signum, signal.SIG_IGN)

    @classmethod
    def getInterpreterArgs(cls, test, interpreter):
        args = plugins.splitcmd(interpreter)
        if len(args) > 0 and args[0] == "ttpython":  # interpreted to mean "whatever python TextTest runs with"
            basename = os.path.basename(sys.executable).lower()
            python = sys.executable if "python" in basename else "python"
            # 'Native launcher' on Windows, such as via Windows installer. Don't know what Python it used
            return [ python, "-u" ] + args[1:]
        else:
            return args

    def quoteLocalArg(self, arg):
        return arg if "$" in arg else pipes.quote(arg)

    def writeScriptLine(self, scriptFile, line):
        scriptFile.write(line.encode(getpreferredencoding()) + b"\n")

    def getRemoteExecuteCmdArgs(self, test, runMachine, localArgs, postfix):
        scriptFileName = test.makeTmpFileName("run_test" + postfix + ".sh", forComparison=0)
        scriptFile = open(scriptFileName, "wb")
        self.writeScriptLine(scriptFile, "#!/bin/sh")
        self.writeScriptLine(scriptFile, "")

        # Need to change working directory remotely
        tmpDir = self.getTmpDirectory(test)
        self.writeScriptLine(scriptFile, "cd " + plugins.quote(tmpDir))

        # Must set the environment remotely
        remoteTmp = test.app.getRemoteTmpDirectory()[1]
        for arg, value in self.getEnvironmentArgs(test, remoteTmp, postfix):
            # Two step export process for compatibility with CYGWIN and older versions of 'sh'
            self.writeScriptLine(scriptFile, arg + "=" + value)
            self.writeScriptLine(scriptFile, "export " + arg)
        if test.app.getConfigValue("remote_shell_program") == "ssh":
            # SSH doesn't kill remote processes, create a kill script
            self.writeScriptLine(scriptFile, 'echo "kill $$" > kill_test.sh')
        cmdString = " ".join(map(self.quoteLocalArg, localArgs))
        if remoteTmp:
            cmdString = cmdString.replace(test.app.writeDirectory, remoteTmp)
        self.writeScriptLine(scriptFile, "exec " + cmdString)
        scriptFile.close()
        os.chmod(scriptFileName, 0o775)  # make executable
        remoteTmp = test.app.getRemoteTestTmpDir(test)[1]
        if remoteTmp:
            test.app.copyFileRemotely(scriptFileName, "localhost", remoteTmp, runMachine)
            remoteScript = os.path.join(remoteTmp, os.path.basename(scriptFileName))
            return test.app.getCommandArgsOn(runMachine, [plugins.quote(remoteScript)])
        else:
            return test.app.getCommandArgsOn(runMachine, [plugins.quote(scriptFileName)])

    def getEnvironmentArgs(self, test, remoteTmp, postfix):
        vars = self.getEnvironmentChanges(test, postfix)
        args = []
        localTmpDir = test.app.writeDirectory
        builtinVars = ["TEXTTEST_CHECKOUT", "TEXTTEST_CHECKOUT_NAME", "TEXTTEST_ROOT", "TEXTTEST_LOG_DIR",
                       "TEXTTEST_SANDBOX", "TEXTTEST_SANDBOX_ROOT", "STORYTEXT_HOME_LOCAL"]
        for var, value in vars:
            if var in builtinVars:
                continue
            if remoteTmp:
                remoteValue = value.replace(localTmpDir, remoteTmp)
            else:
                remoteValue = value

            currentValue = os.getenv(var)
            if currentValue:
                remoteValue = remoteValue.replace(currentValue, "${" + var + "}")
                if var == "PATH" and os.name == "nt":
                    # We assume cygwin paths, make sure we use POSIX path separators
                    remoteValue = remoteValue.replace(";", ":")
            remoteValue = plugins.quote(remoteValue)
            args.append((var, remoteValue))
        return args

    def getTmpDirectory(self, test):
        remoteTmp = test.app.getRemoteTestTmpDir(test)[1]
        if remoteTmp:
            return remoteTmp
        else:
            return test.getDirectory(temporary=1)

    def getTimingArgs(self, test, makeDirs):
        machine, remoteTmp = test.app.getRemoteTestTmpDir(test)
        if remoteTmp:
            frameworkDir = os.path.join(remoteTmp, "framework_tmp")
            if makeDirs:
                test.app.ensureRemoteDirExists(machine, frameworkDir)
            perfFile = os.path.join(frameworkDir, "unixperf")
        else:
            perfFile = test.makeTmpFileName("unixperf", forFramework=1)
        return ["time", "-p", "-o", perfFile]

    def getLocalExecuteCmdArgs(self, test, postfix="", makeDirs=True):
        args = []
        if test.app.hasAutomaticCputimeChecking() and test.app.executingOnPerformanceMachine(test):
            args += self.getTimingArgs(test, makeDirs)

        # Don't expand environment if we're running on a different file system
        expandVars = test.app.getRunMachine() == "localhost" or not test.getConfigValue("remote_copy_program")
        for interpreterName, interpreter in list(test.getConfigValue("interpreters", expandVars=expandVars).items()):
            args += self.getInterpreterArgs(test, interpreter)
            args += test.getCommandLineOptions(stem=interpreterName + "_options")
            if postfix:
                args += test.getCommandLineOptions(stem=interpreterName + "_options" + postfix)
            if os.path.basename(interpreter) == "storytext" and test.app.useVirtualDisplay():
                # Don't display storytext editor under a virtual display!
                args.append("-x")
        args += plugins.splitcmd(test.getConfigValue("executable", expandVars=expandVars))
        args += test.getCommandLineOptions(stem="options")
        if postfix:
            args += test.getCommandLineOptions(stem="options" + postfix)
        return args

    def getExecuteCmdArgs(self, test, runMachine, postfix):
        args = self.getLocalExecuteCmdArgs(test, postfix)
        if runMachine == "localhost":
            return args
        else:
            return self.getRemoteExecuteCmdArgs(test, runMachine, args, postfix)

    def makeFile(self, test, name):
        fileName = test.makeTmpFileName(name)
        return open(fileName, "w")

    def getInputFile(self, test, inputStem):
        inputFileName = test.getFileName(inputStem)
        if inputFileName:
            return inputFileName
        else:
            return os.devnull

    def setUpSuite(self, suite):
        self.describe(suite)



#!/usr/bin/env python
# Majority credit to Sorin Sbarnea @ StackOverflow http://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored
import logging

NORMAL = 0
RED = 1
GREEN = 2

initialEmitter = None

# now we patch Python code to add color support to logging.StreamHandler


def add_coloring_to_emit_windows(fn, desired):
    def _set_color(self, code):
        import ctypes
        # Constants from the Windows API
        self.STD_OUTPUT_HANDLE = -11
        hdl = ctypes.windll.kernel32.GetStdHandle(self.STD_OUTPUT_HANDLE)
        ctypes.windll.kernel32.SetConsoleTextAttribute(hdl, code)

    setattr(logging.StreamHandler, '_set_color', _set_color)

    def new(*args):
        # wincon.h
        FOREGROUND_BLUE = 0x0001  # text color contains blue.
        FOREGROUND_GREEN = 0x0002  # text color contains green.
        FOREGROUND_RED = 0x0004  # text color contains red.
        FOREGROUND_INTENSITY = 0x0008  # text color is intensified.
        FOREGROUND_WHITE = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED

        if desired == RED:
            color = FOREGROUND_RED | FOREGROUND_INTENSITY
        elif desired == GREEN:
            color = FOREGROUND_GREEN | FOREGROUND_INTENSITY

        args[0]._set_color(color)  # apply color
        ret = fn(*args)  # allow the logger to do it's thing
        args[0]._set_color(FOREGROUND_WHITE)  # remove color

        return ret
    return new


def add_coloring_to_emit_ansi(fn, desired):
    # add methods we need to the class
    def new(*args):
        if desired == RED:
            color = '\x1b[31m'  # red
        elif desired == GREEN:
            color = '\x1b[32m'  # green
        # wrap the message in the color tags
        args[1].msg = color + args[1].msg + '\x1b[0m'  # always end in normal

        return fn(*args)
    return new


import platform


def enableOutputColor(color):
    global initialEmitter
    initialEmitter = logging.StreamHandler.emit
    if platform.system() == 'Windows':
        # Windows does not support ANSI escapes and we are using API calls to set the console color
        logging.StreamHandler.emit = add_coloring_to_emit_windows(logging.StreamHandler.emit, color)
    else:
        # all non-Windows platforms are supporting ANSI escapes so we use them
        logging.StreamHandler.emit = add_coloring_to_emit_ansi(logging.StreamHandler.emit, color)


def disableOutputColor():
    global initialEmitter
    logging.StreamHandler.emit = initialEmitter




from gi.repository import Gtk

# Semi-ugly hack to get hold of the button in a treeview header (treeviewcolumn)
# See e.g. http://www.tenslashsix.com/?p=109 or http://piman.livejournal.com/361173.html
# or google on e.g. 'gtk treeview header popup menu' for more info.


class ButtonedTreeViewColumn(Gtk.TreeViewColumn):
    def __init__(self, title="", *args, **kwargs):
        super(ButtonedTreeViewColumn, self).__init__(None, *args, **kwargs)
        label = Gtk.Label(label=title)
        self.set_widget(label)
        label.show()

    def get_button(self):
        return self.get_widget().get_ancestor(Gtk.Button)

    def get_title(self):
        return self.get_widget().get_text()




"""
Module containing code that's only run in the slave jobs when running with a grid engine / queue system
"""

import os
import sys
import time
import socket
import signal
import logging
from .utils import *
from texttestlib import plugins
from texttestlib.default.runtest import RunTest
from texttestlib.default.sandbox import FindExecutionHosts, MachineInfoFinder
from texttestlib.default.actionrunner import ActionRunner
from texttestlib.utils import getUserName
from pickle import dumps
from locale import getpreferredencoding


def importAndCallFromQueueSystem(app, *args):
    moduleName = "queuesystem." + queueSystemName(app).lower()
    return plugins.importAndCall(moduleName, *args)


# Use a non-monitoring runTest, but the rest from unix
class RunTestInSlave(RunTest):
    def getBriefText(self, execMachines):
        return "RUN (" + ",".join(execMachines) + ")"

    def getKillInfoOtherSignal(self, test):
        if os.name == "posix":
            if self.killSignal == signal.SIGUSR1:
                return self.getUserSignalKillInfo(test, "1")
            elif self.killSignal == signal.SIGUSR2:
                return self.getUserSignalKillInfo(test, "2")

        return RunTest.getKillInfoOtherSignal(self, test)

    def getUserSignalKillInfo(self, test, userSignalNumber):
        return importAndCallFromQueueSystem(test.app, "getUserSignalKillInfo", userSignalNumber, self.getExplicitKillInfo)

# Redirect the log mechanism locally in the slave
# Workaround for NFS slowness, essentially: we can't create these files in the master process in case they
# don't propagate in time


class RedirectLogResponder(plugins.Responder):
    done = False

    def notifyAdd(self, test, *args, **kw):
        if not self.done:
            RedirectLogResponder.done = True
            logDir = os.path.join(test.app.writeDirectory, "slavelogs")
            plugins.ensureDirectoryExists(logDir)
            os.chdir(logDir)
            logFile, errFile = test.app.getSubmissionRules(test).getJobFiles()
            errPath = os.path.join(logDir, errFile)
            logPath = os.path.join(logDir, logFile)
            if not os.path.isfile(errPath) and not os.path.isfile(logPath):
                sys.stderr = open(errPath, "w")
                # This is more or less hardcoded to use a timed formatter and write to the file
                handler = logging.FileHandler(logPath)
                formatter = logging.Formatter("%(asctime)s - %(message)s")
                handler.setFormatter(formatter)
                plugins.log.addHandler(handler)


class SocketResponder(plugins.Responder, plugins.Observable):
    synchFiles = False

    def __init__(self, optionMap, *args):
        plugins.Responder.__init__(self)
        plugins.Observable.__init__(self)
        self.killed = False
        self.transferAll = optionMap.get("keepslave") or optionMap.get("keeptmp")
        self.testsForRerun = []
        self.serverAddress = self.getServerAddress(optionMap)

    def getServerAddress(self, optionMap):
        servAddrStr = optionMap.get("servaddr", os.getenv("CAPTUREMOCK_SERVER"))
        if not servAddrStr:
            raise plugins.TextTestError("Cannot run slave, no server address has been provided to send results to!")
        host, port = servAddrStr.split(":")
        return host, int(port)

    def connect(self, sendSocket):
        attempts = 5
        for attempt in range(attempts):
            try:
                sendSocket.connect(self.serverAddress)
                return True
            except socket.error:
                time.sleep(1)
                if attempt == attempts -1:
                    sys.stderr.write("Failed to connect to " + repr(self.serverAddress) + " : " + self.exceptionOutput())
        return False

    def exceptionOutput(self):
        exctype, value = sys.exc_info()[:2]
        from traceback import format_exception_only
        return "".join(format_exception_only(exctype, value))

    def notifyKillProcesses(self, *args):
        self.killed = True

    def getProcessIdentifier(self, test, sendFiles):
        identifier = str(os.getpid())
        rerun = test in self.testsForRerun
        if rerun:
            self.testsForRerun.remove(test)
        return makeIdentifierLine(identifier, sendFiles, False, self.killed, rerun)

    def notifyRerun(self, test):
        self.testsForRerun.append(test)

    def notifyLifecycleChange(self, test, state, changeDesc):
        testData = socketSerialise(test)
        protocol = int(os.getenv("TEXTTEST_PICKLE_PROTOCOL", 2)) # Which pickle protocol to use. Useful to set to plain text for self-tests.
        pickleData = dumps(state, protocol=protocol)
        sendFiles = self.synchFiles and changeDesc == "complete" and (self.transferAll or not test.state.hasSucceeded())
        fullData = self.getProcessIdentifier(test, sendFiles) + os.linesep + testData + os.linesep
        if sendFiles:
            fullData += directorySerialise(test.writeDirectory) + os.linesep
        fullDataBytes = fullData.encode(getpreferredencoding()) + pickleData
        return self.sendAndInterpret(fullDataBytes, self.interpretResponse, state)

    def sendAndInterpret(self, fullData, responseMethod, *args):
        sleepTime = 1
        for _ in range(9):
            sendSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if not self.connect(sendSocket):
                return self.notify("NoMoreExtraTests")
            try:
                response = self.sendData(sendSocket, fullData)
                return responseMethod(response, *args) if responseMethod else True
            except socket.error as e:
                plugins.log.info("Failed to communicate with master process - waiting " +
                                 str(sleepTime) + " seconds and then trying again.")
                plugins.log.info("Error received was " + str(e))
                time.sleep(sleepTime)
                sleepTime *= 2

        message = "Terminating as failed to communicate with master process : " + self.exceptionOutput()
        sys.stderr.write(message)
        plugins.log.info(message.strip())
        self.notify("NoMoreExtraTests")

    def sendData(self, sendSocket, fullData):
        sendSocket.sendall(fullData)
        sendSocket.shutdown(socket.SHUT_WR)
        if self.synchFiles:
            # Remote socket, possibly firewalls that kill connections, possibly other things. Use timeout and be prepared to retry...
            # TCP timeout is typically 30 seconds, give up a bit before that
            sendSocket.settimeout(25)
        response = sendSocket.makefile('rb').read()
        sendSocket.close()
        return str(response, getpreferredencoding())

    def interpretResponse(self, response, state):
        if len(response) > 0:
            appDesc, testPath = socketParse(response)
            appParts = appDesc.split(".")
            self.notify("ExtraTest", testPath, appParts[0], appParts[1:])
        elif state.isComplete():
            self.notify("NoMoreExtraTests")

    def notifyRequiredTestData(self, test, paths):
        if self.synchFiles:
            for path in paths:
                plugins.log.info(test.getIndent() + "Fetching required test data at " + repr(path) + " ...")
            data = makeIdentifierLine(str(os.getpid()), getFiles=True) + "\n" + socketSerialise(test) + "\n" + \
                getUserName() + "@" + getIPAddress([test]) + "\n" + "\n".join(paths)
            self.sendAndInterpret(data.encode(getpreferredencoding()), None)  # Just wait, no response to interpret


class SlaveActionRunner(ActionRunner):
    def notifyAllRead(self, goodSuites):
        # don't ordinarily add a terminator, we might get given more tests via the socket (code above)
        # Need to add one if we haven't found any tests though
        if len(goodSuites) == 0:
            ActionRunner.notifyAllRead(self, goodSuites)

    def notifyRerun(self, *args):
        pass  # don't rerun directly in the slave, tell the master and give it a chance to send the job elsewhere

    def notifyNoMoreExtraTests(self):
        self.diag.info("No more extra tests, adding terminator")
        self.testQueue.put(None)


class FindExecutionHostsInSlave(FindExecutionHosts):
    def getExecutionMachines(self, test):
        return importAndCallFromQueueSystem(test.app, "getExecutionMachines")


class SlaveMachineInfoFinder(MachineInfoFinder):
    def __init__(self):
        self.queueMachineInfo = None

    def findPerformanceMachines(self, test, fileStem):
        perfMachines = []
        resources = test.getCompositeConfigValue("performance_test_resource", fileStem)
        for resource in resources:
            perfMachines += plugins.retryOnInterrupt(self.queueMachineInfo.findResourceMachines, resource)

        rawPerfMachines = MachineInfoFinder.findPerformanceMachines(self, test, fileStem)
        for machine in rawPerfMachines:
            if machine != "any":
                perfMachines += self.queueMachineInfo.findActualMachines(machine)
        if "any" in rawPerfMachines and len(resources) == 0:
            return rawPerfMachines
        else:
            return perfMachines

    def setUpApplication(self, app):
        MachineInfoFinder.setUpApplication(self, app)
        self.queueMachineInfo = importAndCallFromQueueSystem(app, "MachineInfo")

    def getMachineInformation(self, test):
        # Try and write some information about what's happening on the machine
        info = ""
        for machine in test.state.executionHosts:
            for jobLine in self.findRunningJobs(machine):
                info += jobLine + "\n"
        return info

    def findRunningJobs(self, machine):
        return plugins.retryOnInterrupt(self._findRunningJobs, machine)

    def _findRunningJobs(self, machine):
        # On a multi-processor machine performance can be affected by jobs on other processors,
        # as for example a process can hog the memory bus. Describe these so the user can judge
        # for himself if performance is likely to be affected...
        jobsFromQueue = self.queueMachineInfo.findRunningJobs(machine)
        jobs = []
        for user, jobId, jobName in jobsFromQueue:
            jobs.append("Also on " + machine + " : " + user + "'s job " + jobId + " '" + jobName + "'")
        return jobs



import os
import sys
import re
import hashlib
import time
import codecs
import locale
from xml.dom.minidom import parse
from collections import OrderedDict
from glob import glob
from pprint import pprint
from xml.parsers.expat import ExpatError
from locale import getpreferredencoding


class AbortedException(RuntimeError):
    pass


class JobStillRunningException(RuntimeError):
    pass


class FingerprintNotReadyException(RuntimeError):
    pass


def getBuildsDir(jobRoot, jobName):
    projectDir = os.path.join(jobRoot, jobName)
    local = os.path.join(projectDir, "builds")
    if os.path.isdir(local):
        return local

    if hasattr(os, "readlink"):
        link = os.path.join(projectDir, "lastStable")
        try:
            target = os.readlink(link)
            return os.path.dirname(target)
        except OSError:
            return


class BuildDocument:
    versionRegex = re.compile("[0-9]+(\\.[0-9]+)+")
    versionRegexRpm = re.compile("[0-9]+(\\.[0-9]+)+.*.rpm")

    @classmethod
    def create(cls, buildsDir, buildName):
        xmlFile = os.path.join(buildsDir, buildName, "build.xml")
        if os.path.isfile(xmlFile):
            # Error handling due to parsing problems caused last power loss
            try:
                return cls(xmlFile)
            except Exception as e:
                print("WARNING: Error while parsing XML file:" + xmlFile, e)

    def __init__(self, xmlFile):
        self.document = parse(xmlFile)

    def fingerprintStrings(self):
        for obj in self.document.getElementsByTagName("hudson.tasks.Fingerprinter_-FingerprintAction"):
            for entry in obj.getElementsByTagName("string"):
                yield entry.childNodes[0].nodeValue

    def getResult(self):
        for entry in self.document.getElementsByTagName("result"):
            return entry.childNodes[0].nodeValue

    def checkHashes(self, oldHashes, newHashes):
        for currString in self.fingerprintStrings():
            if currString in oldHashes:
                return True, False
            elif currString in newHashes:
                return True, True
        return False, False

    def getArtefactVersion(self, artefactRegex):
        for currString in self.fingerprintStrings():
            if artefactRegex.match(currString):
                versionMatch = self.versionRegex.search(currString)
                if versionMatch:
                    return versionMatch.group(0)

    def hasSourceCodeChange(self):
        for obj in self.document.getElementsByTagName("hudson.model.CauseAction"):
            for causes in obj.getElementsByTagName("causes"):
                return any(("SCMTrigger" in childNode.nodeName for childNode in causes.childNodes))

    def getFingerprint(self, ignoreArtefact):
        prevString = None
        fingerprint = {}
        for currString in self.fingerprintStrings():
            if prevString:
                if ignoreArtefact not in prevString:
                    vregex = self.versionRegexRpm if prevString.endswith(".rpm") else self.versionRegex
                    match = vregex.search(prevString)
                    regex = prevString
                    if match:
                        regex = prevString.replace(match.group(0), vregex.pattern) + "$"
                    fingerprint[regex] = currString, prevString
                prevString = None
            else:
                prevString = currString
        return fingerprint


class FingerprintVerifier:
    def __init__(self, fileFinder, cacheDir):
        self.fileFinder = fileFinder
        self.cacheDir = cacheDir

    def getCacheFileName(self, jobName, buildName):
        rootName = "correct_hashes_" + os.getenv("JENKINS_URL").split("/")[2].replace(":", "")
        return os.path.join(self.cacheDir, rootName, jobName, buildName)

    def getCachedFingerprint(self, jobName, buildName):
        cacheFileName = self.getCacheFileName(jobName, buildName)
        if os.path.isfile(cacheFileName):
            return eval(open(cacheFileName).read())

    def writeCache(self, jobName, buildName, updatedHashes):
        cacheFileName = self.getCacheFileName(jobName, buildName)
        from texttestlib import plugins
        plugins.ensureDirExistsForFile(cacheFileName)
        with open(cacheFileName, "w") as f:
            pprint(updatedHashes, f)

    def md5sum(self, filename):
        md5 = hashlib.md5()
        with open(filename, 'rb') as f:
            for chunk in iter(lambda: f.read(128*md5.block_size), b''):
                md5.update(chunk)
        return md5.hexdigest()

    def getCorrectedHash(self, buildsDir, build, f, hash):
        fullFileFinder = os.path.join(buildsDir, build, self.fileFinder)
        filePattern = f.split(":")[-1].replace("-", "?")
        paths = glob(os.path.join(fullFileFinder, filePattern))
        if len(paths):
            path = paths[0]
            correctHash = self.md5sum(path)
            if correctHash != hash:
                return correctHash


class FingerprintDifferenceFinder:
    def __init__(self, jobRoot, fileFinder, cacheDir):
        self.jobRoot = jobRoot
        self.verifier = FingerprintVerifier(fileFinder, cacheDir) if fileFinder else None

    def findDifferences(self, jobName, build1, build2):
        buildsDir = getBuildsDir(self.jobRoot, jobName)
        if not buildsDir:
            return [], False
        fingerprint1 = self.getFingerprint(buildsDir, jobName, build1)
        fingerprint2 = self.getAndWaitForFingerprint(buildsDir, jobName, build2)
        if not fingerprint1 or not fingerprint2:
            return [], bool(fingerprint2)
        differences = []
        updatedHashes = {}
        for artefact, value in list(fingerprint2.items()):
            if isinstance(value, tuple):
                hash2 = value[0]
                file2 = value[1]
            else:
                hash2 = value
                file2 = artefact
            hash1 = fingerprint1.get(artefact)
            if isinstance(hash1, tuple):
                hash1 = hash1[0]
            if hash1 != hash2:
                if self.verifier:
                    correctedHash = self.verifier.getCorrectedHash(buildsDir, build2, file2, hash2)
                    if correctedHash:
                        hash2 = correctedHash
                        updatedHashes[artefact] = hash2
                    if hash1 == hash2:
                        continue
                differences.append((artefact, hash1, hash2))

        if updatedHashes:
            print("WARNING: incorrect hashes found!")
            print("This is probably due to fingerprint data being wrongly updated from artefacts produced during the build")
            print("Storing a cached file of corrected versions. The following were changed:")
            for artefact, hash in list(updatedHashes.items()):
                print(artefact, fingerprint2.get(artefact)[0], hash)

            for artefact, (hash2, file2) in list(fingerprint2.items()):
                if artefact not in updatedHashes:
                    updatedHashes[artefact] = hash2
            self.verifier.writeCache(jobName, build2, updatedHashes)

        differences.sort()
        return differences, True

    def getAndWaitForFingerprint(self, *args):
        for i in range(500):
            try:
                return self.getFingerprint(*args)
            except FingerprintNotReadyException:
                if i % 10 == 0:
                    print("No Jenkins fingerprints available yet, sleeping...")
                time.sleep(1)

        print("Giving up waiting for fingerprints.")
        raise JobStillRunningException()

    def getFingerprint(self, buildsDir, jobName, buildName):
        if self.verifier:
            cached = self.verifier.getCachedFingerprint(jobName, buildName)
            if cached:
                return cached

        document = BuildDocument.create(buildsDir, buildName)
        fingerprint = document.getFingerprint(jobName) if document is not None else {}
        if not fingerprint:
            result = document.getResult() if document is not None else None
            if result is None and os.getenv("BUILD_NUMBER") == buildName and os.getenv("JOB_NAME") == jobName:
                if os.getenv("BUILD_ID") == "none":
                    # Needed to prevent Jenkins from killing background jobs running after the job has exited
                    # If we have this, we should wait a bit
                    raise FingerprintNotReadyException()
                else:
                    raise JobStillRunningException()
            # No result means aborted (hard) if we're checking a previous run, otherwise it means we haven't finished yet
            elif result == "ABORTED" or (result is None and document is not None):
                raise AbortedException("Aborted in Jenkins")
        return fingerprint


class ChangeSetFinder:
    def __init__(self, jobRoot, jenkinsUrl, bugSystemData):
        self.jobRoot = jobRoot
        self.jenkinsUrl = jenkinsUrl
        self.bugSystemData = bugSystemData

    def parseXmlChangeLog(self, document):
        authors = []
        bugs = []
        for changeset in document.getElementsByTagName("changeset"):
            author = self.parseAuthor(changeset.getAttribute("author"))
            if author and (author not in authors):
                authors.append(author)
            for msgNode in changeset.getElementsByTagName("msg"):
                if len(msgNode.childNodes):
                    msg = msgNode.childNodes[0].nodeValue
                    self.addUnique(bugs, self.getBugs(msg))
        return authors, bugs

    def parsePlainChangeLog(self, fileName):
        authors = []
        bugs = []
        # If there are non-ascii characters in the changelog, there
        # will be encoding errors if the file is opened with plain
        # open(). People's names are expected to be spelled as
        # intended in this day and age. It's not clear why this is
        # needed, a standalone python program reading the same file
        # have no issues.
        with open(fileName) as changelog:
            for line in changelog:
                if "<" in line and "@" in line:
                    author = self.parseAuthor(line.split(": ")[-1])
                    if author and (author not in authors):
                        authors.append(author)
                if line.startswith(" "):
                    self.addUnique(bugs, self.getBugs(line))
        return authors, bugs

    def parseChangeLog(self, xmlFile):
        try:
            document = parse(xmlFile)
        except:
            return self.parsePlainChangeLog(xmlFile)

        return self.parseXmlChangeLog(document)

    def getChangeSetData(self, projectChanges):
        changes = []
        for project, build in projectChanges:
            buildsDir = getBuildsDir(self.jobRoot, project)
            if buildsDir is None:
                continue
            xmlFile = os.path.join(buildsDir, build, "changelog.xml")
            if os.path.isfile(xmlFile):
                authors, bugs = self.parseChangeLog(xmlFile)
                if authors:
                    fullUrl = "/".join([self.jenkinsUrl.rstrip("/"), "job", project, build, "changes"])
                    changes.append((",".join(authors), fullUrl, bugs))
        return changes

    def parseAuthor(self, author):
        if isinstance(author, bytes):
            author = str(author, locale.getpreferredencoding())

        withoutEmail = author.split("<")[0].strip().split("@")[0]
        if "." in withoutEmail:
            return " ".join([part.capitalize() for part in withoutEmail.split(".")])
        else:
            try:
                withoutEmail = withoutEmail.encode("ascii", "xmlcharrefreplace")
            except UnicodeDecodeError as exception:
                print("FAILED to encode name '" + withoutEmail + "' (repr: " + repr(withoutEmail) + ", "
                      + str(type(withoutEmail)) + ", default encoding: '"
                      + sys.getdefaultencoding() + "', default locale: "
                      + str(locale.getdefaultlocale()) + ") due to:\n",
                      exception, "\nIgnoring this entry")
                return None
            return str(withoutEmail, locale.getpreferredencoding())

    def addUnique(self, items, newItems):
        for newItem in newItems:
            if newItem not in items:
                items.append(newItem)

    def getBugs(self, msg):
        bugs = []
        for systemName, location in list(self.bugSystemData.items()):
            try:
                namespace = {}
                exec("from texttestlib.default.knownbugs." + systemName + " import getBugsFromText", globals(), namespace)
                self.addUnique(bugs, namespace["getBugsFromText"](msg, location))
            except ImportError:
                pass
        return bugs


class ProjectData:
    def __init__(self, jobRoot):
        self.data = {}
        workspaceRoot = os.path.dirname(os.getenv("WORKSPACE"))
        for jobName in os.listdir(workspaceRoot):
            jobDir = os.path.join(jobRoot, jobName)
            if os.path.isdir(jobDir):
                subdir = self.getSubdirectory(jobDir)
                workspaceDir = os.path.join(workspaceRoot, jobName, subdir)
                for artefactName, providedScope in self.getArtefactsFromPomFiles(workspaceDir):
                    self.data.setdefault(artefactName, []).append((jobName, providedScope))

    def isAttachedRpm(self, pluginNode):
        return any((goalNode.childNodes[0].nodeValue == "rpm-maven-plugin" for goalNode in pluginNode.getElementsByTagName("artifactId")))

    def getRpmName(self, node):
        for pluginNode in node.getElementsByTagName("plugin"):
            if self.isAttachedRpm(pluginNode):
                for confNode in pluginNode.childNodes:
                    if confNode.nodeName == "configuration":
                        for nameNode in confNode.childNodes:
                            if nameNode.nodeName == "name":
                                return nameNode.childNodes[0].nodeValue

    def getPomData(self, pomFile):
        document = parse(pomFile)
        artifactId, groupId = None, None
        rpmName = None
        artefacts, modules = [], []
        for node in document.documentElement.childNodes:
            if artifactId is None and node.nodeName == "artifactId":
                artifactId = node.childNodes[0].nodeValue
            elif groupId is None and node.nodeName == "groupId":
                groupId = node.childNodes[0].nodeValue
            elif node.nodeName == "modules":
                for subNode in node.childNodes:
                    if subNode.childNodes:
                        modules.append(subNode.childNodes[0].nodeValue)
            elif node.nodeName == "build":
                rpmName = self.getRpmName(node)
        providedScope = any((node.childNodes[0].nodeValue ==
                             "provided" for node in document.getElementsByTagName("scope")))
        groupPrefix = groupId + ":" if groupId else ""
        artefacts.append((groupPrefix + artifactId, providedScope))
        if rpmName:
            artefacts.append((groupPrefix + rpmName, True))
        return artefacts, modules

    def getArtefactsFromPomFiles(self, workspaceDir):
        pomFile = os.path.join(workspaceDir, "pom.xml")
        if not os.path.isfile(pomFile):
            return []

        artefacts, modules = self.getPomData(pomFile)
        for module in modules:
            moduleDir = os.path.join(workspaceDir, module)
            artefacts += self.getArtefactsFromPomFiles(moduleDir)
        return artefacts

    def getSubdirectory(self, jobDir):
        configFile = os.path.join(jobDir, "config.xml")
        if not os.path.isfile(configFile):
            return ""

        try:
            document = parse(configFile)
            for subDir in document.getElementsByTagName("subdir"):
                return subDir.childNodes[0].nodeValue
        except ExpatError as exception:
            print("WARNING: Corrupt config file:\n ", os.path.abspath(
                configFile), "\n Collection of Jenkins data will be incomplete.")
        return ""

    def getProjects(self, artefact):
        currProjArtefact = None
        currProjects = []
        for projArtefact, projects in list(self.data.items()):
            if currProjArtefact is None or len(projArtefact) > len(currProjArtefact):
                if artefact.startswith(projArtefact):
                    currProjArtefact = artefact
                    currProjects = projects
                elif ":" in projArtefact:
                    group, local = projArtefact.split(":", 1)
                    if artefact.startswith(local):
                        currProjArtefact = group + ":" + artefact
                        currProjects = projects

        return currProjArtefact or artefact, currProjects


class ChangeFinder:
    def __init__(self, bugSystemData, markedArtefacts, *args):
        self.jobRoot = os.path.join(os.getenv("JENKINS_HOME"), "jobs")
        self.jobName = os.getenv("JOB_NAME")
        self.projectData = ProjectData(self.jobRoot)
        self.markedArtefacts = markedArtefacts
        self.changeSetFinder = ChangeSetFinder(self.jobRoot, os.getenv("JENKINS_URL"), bugSystemData)
        self.diffFinder = FingerprintDifferenceFinder(self.jobRoot, *args)
        self.handledProjects = []
    def findChanges(self, build1, build2):
        try:
            markedChanges, projectChanges, fingerprintsFound = self.getChangesRecursively(self.jobName, build1, build2)
            if not fingerprintsFound:
                print("WARNING: tried to find Jenkins changes, but no fingerprints found for", self.jobName, "build", build2)
                if os.getenv("BUILD_ID") == "none" and os.getenv("BUILD_NUMBER") == build2:
                    print("(build is ongoing, so should have waited for results)")
        except AbortedException as e:
            # If it was aborted, say this
            return [(str(e), "", [])]

        if self.hasSourceCodeChange(build2):  # If we're polling things from SCM, we should include the changes there
            projectChanges.append((self.jobName, build2))

        # Extract the changeset information from them
        changesFromProjects = self.changeSetFinder.getChangeSetData(projectChanges)
        changesFromMarking = [self.getMarkChangeText(artefact, projectName, build1, build2)
                              for artefact, projectName in markedChanges]
        return changesFromMarking + changesFromProjects

    def getChangesRecursively(self, jobName, build1, build2):
        # Find what artefacts have changed between times build
        differences, fingerprintsFound = self.diffFinder.findDifferences(jobName, build1, build2)
        # Organise them by project
        markedChanges, differencesByProject = self.organiseByProject(differences)
        # For each project, find out which builds were affected
        projectChanges, recursiveChanges = self.getProjectChanges(differencesByProject)
        self.handledProjects.append(jobName)
        for subProj, subBuild1, subBuild2 in recursiveChanges:
            if not subProj in self.handledProjects:
                subMarkedChanges, subProjectChanges, _ = self.getChangesRecursively(subProj, subBuild1, subBuild2)
                for subMarkChange in subMarkedChanges:
                    if subMarkChange not in markedChanges:
                        markedChanges.append(subMarkChange)
                for subProjectChange in subProjectChanges:
                    if subProjectChange not in projectChanges:
                        projectChanges.append(subProjectChange)
        return markedChanges, projectChanges, fingerprintsFound

    def organiseByProject(self, differences):
        differencesByProject = OrderedDict()
        changes = []
        for artefact, oldHash, hash in differences:
            actualArtefact, projects = self.projectData.getProjects(artefact)
            if projects:
                for project, scopeProvided in projects:
                    differencesByProject.setdefault(project, []).append((actualArtefact, oldHash, hash, scopeProvided))
                    if project in self.markedArtefacts:
                        changes.append((actualArtefact, project))
            else:
                projectName = artefact.split(":")[-1].split("[")[0][:-1]
                if projectName in self.markedArtefacts:
                    changes.append((actualArtefact, projectName))

        return changes, differencesByProject

    def getProjectChanges(self, differencesByProject):
        projectChanges = []
        recursiveChanges = []
        for project, diffs in list(differencesByProject.items()):
            buildsDir = getBuildsDir(self.jobRoot, project)
            if buildsDir is None:
                continue
            allBuilds = sorted([build for build in os.listdir(buildsDir) if build.isdigit()], key=lambda b: -int(b))
            oldHashes = [oldHash for _, oldHash, _, _ in diffs]
            newHashes = [hash for _, _, hash, _ in diffs]
            scopeProvided = any((s for _, _, _, s in diffs))
            activeBuild = None
            for build in allBuilds:
                document = BuildDocument.create(buildsDir, build)
                if not document:
                    continue

                if document.getResult() != "FAILURE":
                    matched, matchedNew = document.checkHashes(oldHashes, newHashes)
                    if matched:
                        if matchedNew:
                            activeBuild = build
                        else:
                            if scopeProvided and activeBuild:
                                recursiveChanges.append((project, build, activeBuild))
                            break
                if activeBuild and (project, build) not in projectChanges:
                    projectChanges.append((project, build))
        return projectChanges, recursiveChanges

    def hasSourceCodeChange(self, build):
        buildsDir = getBuildsDir(self.jobRoot, self.jobName)
        if buildsDir:
            doc = BuildDocument.create(buildsDir, build)
            if doc:
                return doc.hasSourceCodeChange()
        return False

    def getMarkChangeText(self, artefact, projectName, build1, build2):
        buildsDir = getBuildsDir(self.jobRoot, self.jobName)
        regex = re.compile(artefact)
        version1 = BuildDocument.create(buildsDir, build1).getArtefactVersion(regex)
        version2 = BuildDocument.create(buildsDir, build2).getArtefactVersion(regex)
        projectNameForDisplay = os.path.basename(projectName)
        if version1 == version2:
            return projectNameForDisplay + " was updated", "", []
        elif version2:
            return projectNameForDisplay + " " + version2, "", []
        else:
            print("WARNING: Artefact version not found for: " + projectName + " build: " + build2)
            return "Artefact version not found", "", []


def getChanges(build1, build2, *args):
    finder = ChangeFinder(*args)
    return finder.findChanges(build1, build2)


def getTimestamp(build):
    if hasattr(os, "readlink"):
        jobRoot = os.path.join(os.getenv("JENKINS_HOME"), "jobs")
        buildsDir = getBuildsDir(jobRoot, os.getenv("JOB_NAME"))
        if buildsDir:
            buildLink = os.path.join(buildsDir, build)
            if os.path.islink(buildLink):
                return os.readlink(buildLink)


def parseEnvAsList(varName):
    if varName in os.environ:
        return os.getenv(varName).split(",")
    else:
        return []


def parseEnvAsDict(varName):
    ret = {}
    for pairText in parseEnvAsList(varName):
        var, value = pairText.split("=")
        ret[var] = value
    return ret


if __name__ == "__main__":
    localFile = os.path.abspath(__file__)  # <root>/texttestlib/default/batch/jenkinschanges.py
    rootDir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(localFile))))
    sys.path.insert(0, rootDir)
    buildName = sys.argv[1]
    if len(sys.argv) > 2:
        prevBuildName = sys.argv[2]
    else:
        prevBuildName = str(int(buildName) - 1)
    pprint(getChanges(prevBuildName, buildName, parseEnvAsDict("BUG_SYSTEM_DATA"), parseEnvAsList("MARKED_ARTEFACTS"),
                      os.getenv("FILE_FINDER", ""), os.getenv("CACHE_DIRECTORY", os.getcwd())))




import sys
import os


def trySetupCoverage():  # pragma: no cover - can hardly measure coverage here :)
    try:
        import coverage
        coverage.process_startup()  # doesn't do anything unless COVERAGE_PROCESS_START is set
    except Exception:
        pass


def loadTestCustomize():
    try:
        # Generic file name to customize the behaviour of Python per test
        import testcustomize
    except ImportError:
        pass


def trySetupCaptureMock():
    try:
        import capturemock
        capturemock.process_startup()  # doesn't do anything unless CAPTUREMOCK_PROCESS_START is set
    except Exception:
        pass


def loadRealSiteCustomize(fileName):  # pragma: no cover - coverage not set up yet
    # must do this before setting up coverage as real sitecustomize might
    # manipulate PYTHONPATH in such a way that coverage can be found
    import imp
    myDir = os.path.dirname(fileName)
    pos = sys.path.index(myDir)
    try:
        file, pathname, description = imp.find_module("sitecustomize", sys.path[pos + 1:])
        if os.path.basename(os.path.dirname(pathname)) == "traffic_intercepts":
            # For the self-tests: don't load another copy ourselves recursively
            loadRealSiteCustomize(pathname)
        else:
            imp.load_module("sitecustomize", file, pathname, description)
    except ImportError:
        pass


loadRealSiteCustomize(__file__)  # pragma: no cover - coverage not set up yet
trySetupCoverage()  # pragma: no cover - coverage not set up yet
loadTestCustomize()  # pragma: no cover - coverage not set up yet
trySetupCaptureMock()  # pragma: no cover - coverage not set up yet



"""
Code to do with the grid engine master process, i.e. submitting slave jobs and waiting for them to report back
"""

import os
import sys
import socket
import signal
import logging
import time
from .utils import *
from queue import Queue
from socketserver import ThreadingTCPServer, StreamRequestHandler
from threading import RLock, Lock
from collections import OrderedDict
from texttestlib import plugins
from texttestlib.default.console import TextDisplayResponder, InteractiveResponder
from texttestlib.default.knownbugs import CheckForBugs
from texttestlib.default.actionrunner import BaseActionRunner
from texttestlib.default.performance import getTestPerformance
from glob import glob
from locale import getpreferredencoding

plugins.addCategory("abandoned", "abandoned", "were abandoned")


class Abandoned(plugins.TestState):
    def __init__(self, freeText):
        plugins.TestState.__init__(self, "abandoned", briefText="job deletion failed",
                                   freeText=freeText, completed=1, lifecycleChange="complete")


class Pending(plugins.TestState):
    defaultBriefText = "PEND"

    def __init__(self, freeText, briefText=None, lifecycleChange="become pending"):
        briefText = briefText or self.defaultBriefText
        plugins.TestState.__init__(self, "pending", freeText=freeText,
                                   briefText=briefText, lifecycleChange=lifecycleChange)

    def makeModifiedState(self, newRunStatus, newDetails, lifecycleChange):
        if newRunStatus != self.briefText:
            newFreeText = self.freeText + "\n" + newDetails
            return self.__class__(newFreeText, newRunStatus, lifecycleChange)


class QueueSystemServer(BaseActionRunner):
    instance = None

    def __init__(self, optionMap, allApps):
        BaseActionRunner.__init__(self, optionMap, logging.getLogger("Queue System Submit"))
        # queue for putting tests when we couldn't reuse the originals
        self.reuseFailureQueue = Queue()
        self.counterLock = Lock()
        self.testCount = 0
        self.testsSubmitted = 0
        self.maxCapacity = 100000  # infinity, sort of
        self.allApps = allApps
        self.jobs = OrderedDict()
        self.submissionRules = {}
        self.killedJobs = {}
        self.queueSystems = {}
        self.reusedTests = {}
        self.reuseOnly = False
        self.allRead = False
        self.submitAddress = None
        self.createDirectories = False
        self.slaveLogDirs = set()
        self.delayedTestsForAdd = []
        self.remainingForApp = OrderedDict()
        appCapacities = []
        for app in allApps:
            appCapacity = self.maxCapacity
            queueSystem = self.getQueueSystem(app)  # populate cache
            queueCapacity = queueSystem.getCapacity() if queueSystem else None
            # If the slaves run somewhere else, they won't create directories for us
            if queueSystem:
                self.createDirectories |= queueSystem.slavesOnRemoteSystem()
            if queueCapacity is not None and queueCapacity < appCapacity:
                appCapacity = queueCapacity
            if appCapacity > 0:
                configCapacity = app.getConfigValue("queue_system_max_capacity")
                # Accept any config capacity that has been set, i.e. is not the
                # same as default.
                if (configCapacity is not None
                    and configCapacity < self.maxCapacity):
                    appCapacity = configCapacity
            appCapacities.append(appCapacity)
        if all((c == 0 for c in appCapacities)):
            raise plugins.TextTestError(
                "The queue system module is reporting zero capacity.\n"
                "Either you have set 'queue_system_max_capacity' to 0 or "
                "something is uninstalled or unavailable. Exiting.")

        self.maxCapacity = min((c for c in appCapacities if c != 0))
        capacityPerSuite = self.maxCapacity / len(allApps)
        for app in allApps:
            self.remainingForApp[app.name] = capacityPerSuite
        QueueSystemServer.instance = self

    def addSuites(self, suites):
        for suite in suites:
            self.slaveLogDirs.add(suite.app.makeWriteDirectory("slavelogs"))
            plugins.log.info("Using "
                             + queueSystemName(suite.app)
                             + " queues for "
                             + suite.app.description(includeCheckout=True))

    def setSlaveServerAddress(self, address):
        self.submitAddress = os.getenv("CAPTUREMOCK_SERVER", address)
        self.testQueue.put("TextTest slave server started on " + address)

    def addTest(self, test):
        if self.createDirectories:
            test.makeWriteDirectory()
        capacityForApp = self.remainingForApp[test.app.name]
        if capacityForApp > 0:
            self.addTestToQueues(test)
            self.remainingForApp[test.app.name] = capacityForApp - 1
        else:
            if test.app.name == list(self.remainingForApp.keys())[-1]:
                self.addTestToQueues(test)  # For the last app (which may be the only one) there is no point in delaying
            else:
                self.delayedTestsForAdd.append(test)

    def queueTestForRerun(self, test):
        # Clear out the previous job reference, otherwise our grid polling will kill it off
        self.jobs[test] = []
        self.addTestToQueues(test)

    def addTestToQueues(self, test):
        with self.counterLock:
            self.testCount += 1
        queue = self.findQueueForTest(test)
        if queue:
            queue.put(test)

    def addDelayedTests(self):
        for test in self.delayedTestsForAdd:
            self.addTestToQueues(test)
        self.delayedTestsForAdd = []

    def notifyAllRead(self, suites):
        self.addDelayedTests()
        BaseActionRunner.notifyAllRead(self, suites)
        self.allRead = True

    def run(self):  # picked up by core to indicate running in a thread
        try:
            self.runAllTests()
            if len(self.jobs):
                self.diag.info("All jobs submitted, polling the queue system now.")
                if self.canPoll():
                    self.pollQueueSystem()
            self.diag.info("No jobs left to poll, exiting thread")
        except:
            self.diag.info("Submit thread exited with exception!")
            plugins.printException()

    def pollQueueSystem(self):
        # Start by polling after 5 seconds, ever after try every 15
        interval = float(os.getenv("TEXTTEST_QS_POLL_INTERVAL", "0.5"))         # Amount of time to wait between checks for exit/completion when polling grid/cloud
        attempts = int(float(os.getenv("TEXTTEST_QS_POLL_WAIT", "5")) / interval) # Amount of time to wait before initiating polling of grid/cloud
        subsequentAttempts = int(float(os.getenv("TEXTTEST_QS_POLL_SUBSEQUENT_WAIT", "15")) / interval) # Amount of time to wait before subsequent polling of grid/cloud
        if attempts >= 0:
            while True:
                for _ in range(attempts):
                    time.sleep(interval)
                    if self.allComplete:
                        return
                    if self.exited:
                        break
                if not self.exited:
                    self.updateJobStatus()
                attempts = subsequentAttempts
                self.diag.info("Trying to rerun queues " + repr(self.testsSubmitted) +
                               " out of " + repr(self.maxCapacity) + " tests submitted")
                # In case any tests have had reruns triggered since we stopped submitting
                self.runQueue(self.getTestForRun, self.runTest, "rerunning", block=False)

    def canPoll(self):
        queueSystem = self.getQueueSystem(list(self.jobs.keys())[0])
        return queueSystem.supportsPolling()

    def updateJobStatus(self):
        queueSystem = self.getQueueSystem(list(self.jobs.keys())[0])
        statusInfo = queueSystem.getStatusForAllJobs()
        self.diag.info("Got status for all jobs : " + repr(statusInfo))
        if statusInfo is not None:  # queue system not available for some reason
            for test, jobs in list(self.jobs.items()):
                if not test.state.isComplete():
                    for jobId, jobName in jobs:
                        status = statusInfo.get(jobId)
                        if status:
                            # Only do this to test jobs (might make a difference for derived configurations)
                            # Ignore filtering states for now, which have empty 'briefText'.
                            self.updateRunStatus(test, status)
                        elif not status and not self.jobCompleted(test, jobName):
                            # Do this to any jobs
                            self.setSlaveFailed(test, self.jobStarted(test, jobName), True, jobId)

    def updateRunStatus(self, test, status):
        newRunStatus, newExplanation = status
        newState = test.state.makeModifiedState(newRunStatus, newExplanation, "grid status update")
        if newState:
            test.changeState(newState)

    def findQueueForTest(self, test):
        # If we've gone into reuse mode and there are no active tests for reuse, use the "reuse failure queue"
        if self.reuseOnly and self.testsSubmitted == 0:
            self.diag.info("Putting " + test.uniqueName + " in reuse failure queue " + self.remainStr())
            return self.reuseFailureQueue
        else:
            self.diag.info("Putting " + test.uniqueName + " in normal queue " + self.remainStr())
            return self.testQueue

    def handleLocalError(self, test, previouslySubmitted):
        self.handleErrorState(test, previouslySubmitted)
        if (self.testCount == 0 and self.allRead) or (self.reuseOnly and self.testsSubmitted == 0):
            self.diag.info("Submitting terminators after local error")
            self.submitTerminators()

    def submitTerminators(self):
        # snap out of our loop if this was the last one. Rely on others to manage the test queue
        self.reuseFailureQueue.put(None)

    def markTestReuse(self, test, newTest):
        self.jobs[newTest] = self.getJobInfo(test)
        with self.counterLock:
            if self.testCount > 1:
                self.testCount -= 1
                postText = self.remainStr()
            else:
                self.submitTerminators()
                postText = ": submitting terminators as final test" # Don't allow test count to drop to 0 here, can cause race conditions
        self.diag.info("Reusing slave from " + test.uniqueName + " for " + newTest.uniqueName + postText)

    def getTestForReuse(self, test, state, tryReuse, doneRerun):
        # Pick up any test that matches the current one's resource requirements
        if not self.exited:
            if test in self.reusedTests:
                newTest = self.reusedTests.get(test)
                newTestName = newTest.uniqueName if newTest else " no test."
                self.diag.info("Repeating answer: using slave from " + test.uniqueName + " for " + newTestName)
                if newTest:
                    self.markTestReuse(test, newTest)
                    return newTest
            else:
                # Don't allow this to use up the terminator
                newTest = self.getTest(block=False, replaceTerminators=True)
                if newTest:
                    if tryReuse and self.allowReuse(test, state, newTest):
                        if not doneRerun:
                            self.reusedTests[test] = newTest
                        self.markTestReuse(test, newTest)
                        return newTest
                    else:
                        self.diag.info("Adding to reuse failure queue : " + newTest.uniqueName)
                        self.reuseFailureQueue.put(newTest)
                else:
                    self.diag.info("No tests available for reuse : " + test.uniqueName)
                self.reusedTests[test] = None

        # Allowed a submitted job to terminate
        with self.counterLock:
            self.testsSubmitted -= 1
            self.diag.info("No reuse for " + test.uniqueName + " : " +
                           repr(self.testsSubmitted) + " tests still submitted")
            if self.exited and self.testsSubmitted == 0:
                self.diag.info("Forcing termination")
                self.submitTerminators()

    def allowReuse(self, oldTest, oldState, newTest):
        # Don't reuse jobs that have been killed
        if newTest.state.isComplete() or oldState.category == "killed":
            return False

        if oldTest.getConfigValue("queue_system_proxy_executable") or \
           newTest.getConfigValue("queue_system_proxy_executable"):
            return False

        # Jobs maintain the same virtual display instance where possible, if they require different settings they can't be reused
        if oldTest.getConfigValue("virtual_display_extra_args") != newTest.getConfigValue("virtual_display_extra_args") or \
                oldTest.getConfigValue("virtual_display_count") != newTest.getConfigValue("virtual_display_count"):
            return False

        oldRules = self.getSubmissionRules(oldTest)
        newRules = self.getSubmissionRules(newTest)
        return oldRules.allowsReuse(newRules)

    def getJobSubmissionRules(self, test):
        proxyRules = test.app.getProxySubmissionRules(test)
        if proxyRules:
            return proxyRules
        else:
            return self.getSubmissionRules(test)

    def getSubmissionRules(self, test):
        if test in self.submissionRules:
            return self.submissionRules[test]
        else:
            submissionRules = test.app.getSubmissionRules(test)
            self.submissionRules[test] = submissionRules
            return submissionRules

    def getTest(self, block, replaceTerminators=False):
        testOrStatus = self.getItemFromQueue(self.testQueue, block, replaceTerminators)
        if not testOrStatus:
            return
        if type(testOrStatus) == str:
            self.sendServerState(testOrStatus)
            return self.getTest(block)
        else:
            return testOrStatus

    def sendServerState(self, state):
        self.diag.info("Sending server state '" + state + "'")
        mimServAddr = os.getenv("CAPTUREMOCK_SERVER")
        if mimServAddr:
            host, port = mimServAddr.split(":")
            serverAddress = (host, int(port))
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(serverAddress)
            sock.sendall(("SUT_SERVER:" + state + "\n").encode(getpreferredencoding()))
            sock.close()

    def getTestForRunNormalMode(self, block):
        self.reuseOnly = False
        reuseFailure = self.getItemFromQueue(self.reuseFailureQueue, block=False)
        if reuseFailure:
            self.diag.info("Found a reuse failure...")
            return reuseFailure
        else:
            self.diag.info("Waiting for new tests...")
            newTest = self.getTest(block=block)
            if newTest:
                return newTest
            else:
                # Make sure we pick up anything that failed in reuse while we were submitting the final test...
                self.diag.info("No normal test found, checking reuse failures...")
                return self.getItemFromQueue(self.reuseFailureQueue, block=False)

    def getTestForRunReuseOnlyMode(self, block):
        self.reuseOnly = True
        self.diag.info("Waiting for reuse failures...")
        reuseFailure = self.getItemFromQueue(self.reuseFailureQueue, block=block)
        if reuseFailure:
            return reuseFailure
        elif self.testCount > 0 and self.testsSubmitted < self.maxCapacity:
            # Try again, the capacity situation has changed...
            return self.getTestForRunNormalMode(block)

    def getTestForRun(self, block=True):
        if (self.testCount == 0 and self.allRead) or (self.testsSubmitted < self.maxCapacity):
            return self.getTestForRunNormalMode(block)
        elif self.reuseCanFail():
            return self.getTestForRunReuseOnlyMode(block)
        else:
            self.diag.info("No more tests found, reuse cannot fail, stopping.")

    def reuseCanFail(self):
        return any((not qs.slavesOnRemoteSystem() for qs in list(self.queueSystems.values())))

    def notifyAllComplete(self):
        BaseActionRunner.notifyAllComplete(self)
        self.cleanup(final=True)
        if self.reuseOnly: # could still be hanging waiting for this, make sure we terminate
            self.submitTerminators()

        errors = {}
        errorFiles = []
        for logDir in self.slaveLogDirs:
            errorFiles += list(filter(os.path.getsize, glob(os.path.join(logDir, "*.errors"))))
        self.diag.info("All complete, processing " + str(len(errorFiles)) + " error files...")
        if len(errorFiles) == 0:
            return

        for fileName in errorFiles:
            contents = None
            # Take the shortest (i.e. most filtered) one
            for app in self.allApps:
                currContent = app.filterErrorText(fileName)
                if contents is None or len(currContent) < len(contents):
                    contents = currContent
            if contents:
                errors[contents] = os.path.basename(fileName)[:-7]

        for msg, jobName in list(errors.items()):
            sys.stderr.write("WARNING: error produced by slave job '" + jobName + "'\n" + msg)

    def cleanup(self, final=False):
        cleanupComplete = True
        if self.jobs:
            queueSystem = self.getQueueSystem(list(self.jobs.keys())[0])
            cleanupComplete &= queueSystem.cleanup(final)
        if cleanupComplete and not final:
            self.sendServerState("Completed submission of all tests")

    def remainStr(self):
        return " : " + str(self.testCount) + " tests remain, " + str(self.testsSubmitted) + " are submitted."

    def runTest(self, test):
        submissionRules = self.getSubmissionRules(test)
        commandArgs = self.getSlaveCommandArgs(test, submissionRules)
        plugins.log.info("Q: Submitting " + repr(test) + submissionRules.getSubmitSuffix())
        sys.stdout.flush()
        self.jobs[test] = []  # Preliminary jobs aren't interesting any more
        slaveEnv = OrderedDict()
        if not self.submitJob(test, submissionRules, commandArgs, slaveEnv):
            return

        with self.counterLock:
            self.testCount -= 1
            self.testsSubmitted += 1
            self.diag.info("Submission successful" + self.remainStr())
        if not test.state.hasStarted():
            test.changeState(self.getPendingState(test))
        if self.testsSubmitted == self.maxCapacity:
            self.sendServerState("Completed submission of tests up to capacity")

    def fixConfigEnv(self, env, test):
        for envVar in test.getConfigValue("queue_system_environment"):
            val = os.getenv(envVar)
            if val is not None:
                env[envVar] = val

    def fixProxyVar(self, env, test, withProxy):
        if withProxy and test.getConfigValue("queue_system_proxy_executable"):
            env["TEXTTEST_SUBMIT_COMMAND_ARGS"] = "?"

    def fixDisplay(self, env):
        # Must make sure SGE jobs don't get a locally referencing DISPLAY
        display = os.environ.get("DISPLAY")
        if display and display.startswith(":"):
            env["DISPLAY"] = plugins.gethostname() + display

    def getPendingState(self, test):
        return Pending(freeText="Job pending in " + queueSystemName(test.app))

    def getSlaveCommandArgs(self, test, submissionRules):
        queueSystem = self.getQueueSystem(test)
        args = queueSystem.getTextTestArgs()
        if queueSystem.slavesOnRemoteSystem():
            args += ["-home", os.path.expanduser("~")]
        return args + ["-d", ":".join(self.optionMap.rootDirectories),
                       "-a", test.app.name + test.app.versionSuffix(),
                       "-l", "-tp", test.getRelPath()] + \
            self.getSlaveArgs(test) + self.getRunOptions(test.app, submissionRules)

    def getSlaveArgs(self, test):
        return ["-slave", test.app.writeDirectory, "-servaddr", self.submitAddress]

    def getRunOptions(self, app, submissionRules):
        runOptions = []
        for slaveSwitch in app.getSlaveSwitches():
            if slaveSwitch in self.optionMap:
                option = "-" + slaveSwitch
                runOptions.append(option)
                value = self.optionMap.get(slaveSwitch)
                if value:
                    runOptions.append(value)

        if "xs" in self.optionMap:
            runOptions.append("-x")
            runOptions.append("-xr")
            runOptions.append(self.optionMap.get("xr", os.path.expandvars("$TEXTTEST_PERSONAL_LOG/logging.debug")))
            runOptions.append("-xw")
            runOptions.append(os.path.expandvars("$TEXTTEST_PERSONAL_LOG/" + submissionRules.getJobName()))
        return runOptions

    def getSlaveLogDir(self, test):
        return os.path.join(test.app.writeDirectory, "slavelogs")

    def setRemoteProcessId(self, test, pid):
        jobInfo = self.getJobInfo(test)
        if len(jobInfo) > 0:
            # Take the most recent job, it's hopefully the most interesting
            jobId = jobInfo[-1][0]
            queueSystem = self.getQueueSystem(test)
            queueSystem.setRemoteProcessId(jobId, pid)

    def getRemoteTestTmpDir(self, test):
        jobInfo = self.getJobInfo(test)
        if len(jobInfo) > 0:
            # Take the most recent job, it's hopefully the most interesting
            jobId = jobInfo[-1][0]
            queueSystem = self.getQueueSystem(test)
            remoteMachine = queueSystem.getRemoteTestMachine(jobId)
            if remoteMachine:
                return remoteMachine, test.writeDirectory

    def getSubmitCmdArgs(self, test, *args):
        queueSystem = self.getQueueSystem(test)
        return queueSystem.getSubmitCmdArgs(*args)

    def getQueueSystemCommand(self, test):
        submissionRules = self.getSubmissionRules(test)
        cmdArgs = self.getSubmitCmdArgs(test, submissionRules)
        text = queueSystemName(test) + " Command   : " + plugins.commandLineString(cmdArgs) + " ...\n" + \
            "Slave Command : " + " ".join(self.getSlaveCommandArgs(test, submissionRules)) + "\n"
        proxyArgs = self.getProxyCmdArgs(test)
        if proxyArgs:
            return queueSystemName(test) + " Proxy Command   : " + plugins.commandLineString(proxyArgs) + "\n" + text
        else:
            return text

    def getProxyCmdArgs(self, test, slaveEnv={}):
        proxyCmd = test.getConfigValue("queue_system_proxy_executable")
        if proxyCmd:
            proxyOptions = test.getCommandLineOptions("proxy_options")
            fullProxyCmdArgs = [proxyCmd] + proxyOptions
            proxyRules = self.getJobSubmissionRules(test)
            return self.getSubmitCmdArgs(test, proxyRules, fullProxyCmdArgs, slaveEnv)
        else:
            return []

    def modifyCommandForProxy(self, test, cmdArgs, slaveEnv):
        proxyArgs = self.getProxyCmdArgs(test, slaveEnv)
        if proxyArgs:
            cmdArgs[1:1] = ["-sync", "y", "-V"]  # must sychronise in the proxy
            # Proxy likes to set environment variables, make sure they get forwarded
            slaveEnv["TEXTTEST_SUBMIT_COMMAND_ARGS"] = repr(cmdArgs) # Exact command arguments to run TextTest slave, for use by proxy
            return proxyArgs
        else:
            return cmdArgs

    def submitJob(self, test, submissionRules, commandArgs, slaveEnv, withProxy=True, jobType=""):
        self.diag.info("Submitting job at " + plugins.localtime() + ":" + repr(commandArgs))
        self.diag.info("Creating job at " + plugins.localtime())
        self.fixDisplay(slaveEnv)
        self.fixConfigEnv(slaveEnv, test)
        self.fixProxyVar(slaveEnv, test, withProxy)
        cmdArgs = self.getSubmitCmdArgs(test, submissionRules, commandArgs, slaveEnv)
        if withProxy:
            cmdArgs = self.modifyCommandForProxy(test, cmdArgs, slaveEnv)

        jobName = submissionRules.getJobName()
        self.diag.info("Creating job " + jobName + " with command arguments : " + " ".join(cmdArgs))
        with self.lock:
            if self.exited:
                self.cancel(test)
                plugins.log.info("Q: Submission cancelled for " + repr(test) + " - exit underway")
                return False

            self.lockDiag.info("Got lock for submission")
            queueSystem = self.getQueueSystem(test)
            logDir = self.getSlaveLogDir(test)
            jobId, errorMessage = queueSystem.submitSlaveJob(cmdArgs, slaveEnv, logDir, submissionRules, jobType)
            if jobId is not None:
                self.diag.info("Job created with id " + jobId)
                # if the slaves run elsewhere (e.g. the cloud) then the capacity of the system can change dynamically depending on what is available
                if queueSystem.slavesOnRemoteSystem():
                    self.checkQueueCapacity(queueSystem)
                self.jobs.setdefault(test, []).append((jobId, jobName))
                self.lockDiag.info("Releasing lock for submission...")
                return True
            else:
                self.diag.info("Job not created : " + errorMessage)
                test.changeState(plugins.Unrunnable(errorMessage, "NOT SUBMITTED"))
                self.handleErrorState(test)
                return False

    def checkQueueCapacity(self, queueSystem):
        queueCapacity = queueSystem.getCapacity()
        if queueCapacity:
            with self.counterLock:
                if queueCapacity < self.maxCapacity:
                    self.maxCapacity = queueCapacity

    def handleErrorState(self, test, previouslySubmitted=False):
        with self.counterLock:
            if self.maxCapacity > 1:
                self.maxCapacity -= 1
            if previouslySubmitted:
                self.testsSubmitted -= 1
            else:
                self.testCount -= 1
        self.diag.info(repr(test) + " in error state" + self.remainStr())
        bugchecker = CheckForBugs()
        self.setUpSuites(bugchecker, test)
        bugchecker(test)
        test.actionsCompleted()

    def setUpSuites(self, bugchecker, test):
        if test.parent:
            bugchecker.setUpSuite(test.parent)
            self.setUpSuites(bugchecker, test.parent)

    def _getJobFailureInfo(self, test):
        jobInfo = self.getJobInfo(test)
        # Take the most recent job, it's hopefully the most interesting
        jobId = jobInfo[-1][0] if len(jobInfo) > 0 else None
        queueSystem = self.getQueueSystem(test)
        return queueSystem.getJobFailureInfo(jobId)

    def getSlaveErrors(self, test, name):
        errorTexts = [self.getSlaveErrorText(test, jobName, desc)
                      for jobName, desc in self.getErrorJobNames(test, name)]
        return "\n".join([text for text in errorTexts if text])

    def getErrorJobNames(self, test, name):
        jobNames = []
        for _, jobName in self.getJobInfo(test):
            jobNames.append((jobName, name))
            if jobName.startswith("Test-"):
                jobNames.append(("Proxy-" + jobName[5:], name + " Proxy"))
        return jobNames

    def getSlaveErrorText(self, test, jobName, desc):
        errFile = os.path.join(self.getSlaveLogDir(test), jobName + ".errors")
        if os.path.isfile(errFile):
            errors = open(errFile).read()
            if errors:
                return "-" * 10 + " Error messages written by " + desc + " job " + "-" * 10 + \
                    "\n" + errors
        return ""

    def getJobInfo(self, test):
        return self.jobs.get(test, [])

    def killJob(self, test, jobId, jobName):
        prevTest, prevJobExisted = self.killedJobs.get(jobId, (None, False))
        # Killing the same job for other tests should result in the cached result being returned
        if prevTest and test is not prevTest:
            return prevJobExisted

        self.describeJob(test, jobId, jobName)
        queueSystem = self.getQueueSystem(test)
        jobExisted = queueSystem.killJob(jobId)
        self.killedJobs[jobId] = test, jobExisted
        return jobExisted

    def getQueueSystem(self, test):
        queueModuleText = queueSystemName(test)
        if queueModuleText is None:
            return None
        queueModule = queueModuleText.lower()
        if queueModule in self.queueSystems:
            return self.queueSystems[queueModule]

        namespace = {}
        command = "from ." + queueModule + " import QueueSystem as _QueueSystem"
        exec(command, globals(), namespace)
        system = namespace["_QueueSystem"](test)
        self.queueSystems[queueModule] = system
        return system

    def changeState(self, test, newState, previouslySubmitted=True):
        test.changeState(newState)
        self.handleLocalError(test, previouslySubmitted)

    def killTests(self):
        # If we've been killed with some sort of limit signal, wait here until we know
        # all tests terminate. Otherwise we rely on them terminating naturally, and if they don't
        wantStatus = self.killSignal and self.killSignal not in [signal.SIGINT, signal.SIGTERM]
        killedTests = []
        for test, jobList in list(self.jobs.items()):
            if not test.state.isComplete():
                for jobId, jobName in jobList:
                    if self.killTest(test, jobId, jobName, wantStatus):
                        killedTests.append((test, jobId))
        if wantStatus:
            self.waitForKill(killedTests)

    def waitForKill(self, killedTests):
        # Wait for a minute for the kill to take effect, otherwise give up
        stillRunning = killedTests
        for attempt in range(1, 61):
            stillRunning = [test_jobId for test_jobId in stillRunning if not test_jobId[0].state.isComplete()]
            if len(stillRunning) == 0:
                return
            time.sleep(1)
            for test, jobId in stillRunning:
                plugins.log.info("T: Cancellation in progress for " + repr(test) +
                                 ", waited " + str(attempt) + " seconds so far.")
        for test, jobId in stillRunning:
            name = queueSystemName(test.app)
            freeText = "Could not delete test in " + name + " (job " + jobId + "): have abandoned it"
            self.changeState(test, Abandoned(freeText))

    def killOrCancel(self, test):
        # Explicitly chose test to kill (from the GUI)
        jobInfo = self.getJobInfo(test)
        if len(jobInfo) > 0:
            for jobId, jobName in jobInfo:
                self.killTest(test, jobId, jobName, wantStatus=True)
        else:
            self.diag.info("No job info found from queue system server, changing state to cancelled")
            return self.cancel(test)

    def killTest(self, test, jobId, jobName, wantStatus):
        self.diag.info("Killing test " + repr(test) + " in state " + test.state.category)
        jobExisted = self.killJob(test, jobId, jobName)
        startNotified = self.jobStarted(test, jobName)
        if jobExisted:
            if startNotified:
                self.diag.info("Job " + jobId + " was running.")
                return True
            else:
                self.diag.info("Job " + jobId + " was pending.")
                self.setKilledPending(test, jobId)
                return False
        else:
            self.diag.info("Job " + jobId + " did not exist.")
            # might get here when the test completed since we checked...
            if not test.state.isComplete():
                self.setSlaveFailed(test, startNotified, wantStatus, jobId)
        return False

    def setSuspendStateForTests(self, tests, newState):
        for test in tests:
            queueSystem = self.getQueueSystem(test)
            for jobId, _ in self.getJobInfo(test):
                queueSystem.setSuspendState(jobId, newState)

    def jobStarted(self, test, *args):
        return test.state.hasStarted()

    def jobCompleted(self, test, *args):
        return test.state.isComplete()

    def setKilledPending(self, test, jobId):
        timeStr = plugins.localtime("%H:%M")
        briefText = "cancelled pending job at " + timeStr
        freeText = "Test job " + jobId + " was cancelled (while still pending in " + queueSystemName(test.app) +\
                   ") at " + timeStr
        self.cancel(test, briefText, freeText)

    def getJobFailureInfo(self, test, wantStatus):
        if wantStatus:
            return self._getJobFailureInfo(test)
        else:
            # Job accounting info can take ages to find, don't do it from GUI quit
            return "No accounting info found as quitting..."

    def setSlaveFailed(self, test, startNotified, wantStatus, jobId):
        failReason, fullText = self.getSlaveFailure(test, startNotified, wantStatus)
        fullText = failReason + "\nJob ID was " + jobId + "\n" + fullText
        self.changeState(test, self.getSlaveFailureState(startNotified, failReason, fullText))

    def getSlaveFailure(self, test, startNotified, wantStatus):
        fullText = ""
        name = queueSystemName(test.app)
        slaveErrors = self.getSlaveErrors(test, name)
        if slaveErrors:
            fullText += slaveErrors

        fullText += self.getJobFailureInfo(test, wantStatus)
        return self.getSlaveFailureBriefText(name, startNotified), fullText

    def getSlaveFailureBriefText(self, name, startNotified):
        if startNotified:
            return "no report, possibly killed with SIGKILL"
        else:
            return name + " job exited"

    def getSlaveFailureState(self, startNotified, failReason, fullText):
        if startNotified:
            return plugins.TestState("killed", briefText=failReason,
                                     freeText=fullText, completed=1, lifecycleChange="complete")
        else:
            return plugins.Unrunnable(briefText=failReason, freeText=fullText, lifecycleChange="complete")

    def getPostText(self, test, jobId):
        name = queueSystemName(test.app)
        return "in " + name + " (job " + jobId + ")"

    def describeJob(self, test, jobId, *args):
        postText = self.getPostText(test, jobId)
        plugins.log.info("T: Cancelling " + repr(test) + " " + postText)

# Used in slave


class BasicSubmissionRules:
    classPrefix = "Test"

    def __init__(self, test):
        self.test = test
        self.diag = logging.getLogger("Submission Rules")

    def getJobName(self):
        path = self.test.getRelPath()
        parts = path.split(os.sep)
        parts.reverse()
        name = self.classPrefix + "-" + ".".join(parts) + "-" + repr(self.test.app).replace(" ", "_").replace("/", "_")
        return name.replace(":", "_")

    def getJobFiles(self):
        jobName = self.getJobName()
        return jobName + ".log", jobName + ".errors"


class SubmissionRules(BasicSubmissionRules):
    def __init__(self, optionMap, test):
        BasicSubmissionRules.__init__(self, test)
        self.optionMap = optionMap
        self.configResources = self.getConfigResources(test)
        self.processesNeeded = self.getProcessesNeeded()

    def getProcessesNeeded(self):
        return 1

    def getExtraSubmitArgs(self):  # pragma: no cover - documentation only
        return []

    def getParallelEnvironment(self):
        return ""

    def useCoreBinding(self):
        return False

    def findPriority(self):
        return 0

    def findResourceList(self):
        return self.configResources

    def findQueue(self):
        cmdQueue = self.optionMap.get("q", "")
        if cmdQueue:
            return cmdQueue
        configQueue = self.test.app.getConfigValue("default_queue")
        if configQueue != "texttest_default":
            return configQueue

        return self.findDefaultQueue()

    def findDefaultQueue(self):
        return ""

    def findMachineList(self):
        return []

    def forceOnPerformanceMachines(self):
        return False

    def allowsReuse(self, newRules):
        # Don't care about the order of the resources
        return set(self.configResources) == set(newRules.configResources)


class ProxySubmissionRules(SubmissionRules):
    classPrefix = "Proxy"

    def getConfigResources(self, test):
        return test.getConfigValue("queue_system_proxy_resource")


class TestSubmissionRules(SubmissionRules):
    classPrefix = "Test"

    def getConfigResources(self, test):
        if "reconnect" not in self.optionMap:
            configSettings = test.getConfigValue("queue_system_resource")
            envSetting = os.path.expandvars(test.getEnvironment("QUEUE_SYSTEM_RESOURCE", "")) # Deprecated. See "queue_system_resource" in config file docs
            return configSettings + [envSetting] if envSetting else configSettings
        else:
            return ""

    def getProcessesNeeded(self):
        if "reconnect" not in self.optionMap:
            envSetting = self.test.getEnvironment("QUEUE_SYSTEM_PROCESSES", "")  # Deprecated. See "queue_system_processes" in config file docs
            return int(envSetting) if envSetting else self.test.getConfigValue("queue_system_processes")
        else:
            return 1

    def getExtraSubmitArgs(self):
        if "reconnect" not in self.optionMap:
            envSetting = os.path.expandvars(self.test.getEnvironment("QUEUE_SYSTEM_SUBMIT_ARGS", ""))  # Deprecated. See "queue_system_submit_args" in config file docs
            argStr = envSetting or self.test.getConfigValue("queue_system_submit_args")
            return plugins.splitcmd(argStr)
        else:
            return []

    def getSubmitSuffix(self):
        name = queueSystemName(self.test)
        queue = self.findQueue()
        if queue:
            return " to " + name + " queue " + queue
        else:
            return " to default " + name + " queue"

    def getParallelEnvironment(self):
        return self.test.getConfigValue("parallel_environment_name")

    def findResourceList(self):
        resourceList = []
        self.diag.info("Finding resource list for " + repr(self.test))
        cmdResource = self.optionMap.get("R")
        if cmdResource:
            self.diag.info("Adding from command line resource: " + repr(cmdResource))
            resourceList.append(cmdResource)
        resourceList += self.configResources
        machine = self.optionMap.get("m")
        if machine and machine != "localhost":
            # Won't work with LSF, but can't be bothered to figure it out there for now...
            machineResource = "hostname=" + machine
            self.diag.info("Adding from command line machine: " + repr(machineResource))
            resourceList.append(machineResource)
        if self.forceOnPerformanceMachines():
            resources = self.getConfigValue("performance_test_resource")
            self.diag.info(
                "Forcing on performance machines: adding from all keys in performance_test_resource : " + repr(",".join(resources)))
            for resource in resources:
                resourceList.append(resource)
        self.diag.info("Resource list for " + repr(self.test) + " = " + ",".join(resourceList))
        return resourceList

    def getConfigValue(self, configKey):
        configDict = self.test.getConfigValue(configKey)
        defVal = configDict.get("default")
        if len(defVal) > 0:
            return defVal
        for val in list(configDict.values()):
            if len(val) > 0 and val[0] != "any" and val[0] != "none":
                return val
        return []

    def findMachineList(self):
        self.diag.info("Finding machine list for " + repr(self.test))
        if not self.forceOnPerformanceMachines():
            return []
        performanceMachines = self.getConfigValue("performance_test_machine")
        self.diag.info("Machine list for " + repr(self.test) + " = " + ",".join(performanceMachines))
        return performanceMachines

    def getJobFiles(self):
        jobName = self.getJobName()
        return jobName + ".log", jobName + ".errors"

    def forceOnPerformanceMachines(self):
        if "perf" in self.optionMap:
            self.diag.info("-perf flag provided, forcing on performance machines")
            return True

        minTimeForce = plugins.getNumberOfSeconds(str(self.test.getConfigValue("min_time_for_performance_force")))
        if minTimeForce >= 0:
            testPerf = getTestPerformance(self.test)
            if testPerf > minTimeForce:
                self.diag.info("Test expects to take " + repr(testPerf) +
                               " seconds, min_time_for_performance_force = " + repr(minTimeForce) + " seconds")
                return True
        # If we haven't got a log_file yet, we should do this so we collect performance reliably
        logFileStem = self.test.getConfigValue("log_file")
        logFile = self.test.getFileName(logFileStem)
        if logFile is None:
            self.diag.info("No log file (file with stem " + repr(logFileStem) +
                           ") exists in test yet, forcing on performance machines to collect initial files")
        return logFile is None

    def allowsReuse(self, newRules):
        if "reconnect" in self.optionMap:
            return True  # should be able to reconnect anywhere...
        else:
            # Don't care about the order of the resources
            return set(self.findResourceList()) == set(newRules.findResourceList()) and \
                self.processesNeeded == newRules.processesNeeded


class SlaveRequestHandler(StreamRequestHandler):
    def handle(self):
        identifier = str(self.rfile.readline().strip(), getpreferredencoding())
        if identifier == "TERMINATE_SERVER":
            return

        self.handleMessage(identifier)

    def handleMessage(self, identifier):
        # Don't use port, it changes all the time
        identifier, sendFiles, getFiles, tryReuse, rerun = parseIdentifier(identifier)
        testString = str(self.rfile.readline().strip(), getpreferredencoding())
        test = self.server.getTest(testString)
        if test is None:
            clientHost = self.client_address[0]
            sys.stderr.write("WARNING: Received request from hostname " + self.getHostName(clientHost) +
                             " (process " + identifier + ")\nwhich could not be parsed:\n'" + testString + "'\n")
        elif getFiles:
            self.pushFiles(test)
        elif not test.state.isComplete() or not test.state.hasResults():  # we might have killed it already...
            if sendFiles:
                self.server.diag.info("Test " + test.uniqueName +
                                      " - receiving files sent from slave to sandbox directory")
                directoryUnserialise(test.writeDirectory, self.rfile)
            # Don't use port, it changes all the time
            self.handleRequestFromHost(test, identifier, tryReuse, rerun)
        else:
            self.server.diag.info("Test " + test.uniqueName + " already complete, ignoring new results")
            self.sendReuseResponse(test, test.state, tryReuse, False)
        self.connection.shutdown(socket.SHUT_RDWR)

    def getHostName(self, ipAddress):
        try:
            return socket.gethostbyaddr(ipAddress)[0].split(".")[0]
        except socket.error:
            return ipAddress

    def pushFiles(self, test):
        encoding = getpreferredencoding()
        userAndHost = str(self.rfile.readline().strip(), encoding)
        paths = []
        for line in self.rfile:
            paths.append(str(line.strip(), encoding))
        self.server.pushFiles(test, userAndHost, paths)

    def sendReuseResponse(self, *args):
        newTest = QueueSystemServer.instance.getTestForReuse(*args)
        if newTest:
            response = socketSerialise(newTest)
            self.server.diag.info("Sending reuse response " + response)
            self.wfile.write(response.encode(getpreferredencoding()))

    def handleRequestFromHost(self, test, pid, tryReuse, rerun):
        # The updates are only for testing against old slave traffic,
        # a bit sad we can't disable them when not testing...
        _, state = test.getNewState(self.rfile, updatePaths=True)
        if test.state.isComplete():
            state.lifecycleChange = "recalculated"
        doneRerun = self.server.changeStateOrRerun(test, state, rerun)
        try:
            self.connection.shutdown(socket.SHUT_RD)
        except socket.error:
            # This only occurs on a mac, and doesn't affect functionality.
            pass
        if state.isComplete():
            self.sendReuseResponse(test, state, tryReuse, doneRerun)
        else:
            QueueSystemServer.instance.setRemoteProcessId(test, pid)


class SlaveServerResponder(plugins.Responder, ThreadingTCPServer):
    # Python's default value of 5 isn't very much...
    # There doesn't seem to be any disadvantage of allowing a longer queue, so we will use the system's maximum size
    request_queue_size = socket.SOMAXCONN

    def __init__(self, optionMap, allApps):
        plugins.Responder.__init__(self)
        ThreadingTCPServer.__init__(self, (getIPAddress(allApps), 0), self.handlerClass())
        self.testMap = {}
        self.testLocks = {}
        self.filePushLock = Lock()
        self.filePushProcesses = {}
        self.diag = logging.getLogger("Slave Server")
        self.terminate = False
        self.totalReruns = 0
        self.maxReruns = max(filter(lambda x: x is not None,
                                    (app.getBatchConfigValue("queue_system_max_reruns") for app in allApps)))

        # If a client rings in and then the connectivity is lost, we don't want to hang waiting for it forever
        # So we enable keepalive that will check the connection if no data is received for a while
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, True)
        # Default is 2 hours here on Linux which is rather a long time to wait for anything to happen.
        # We give up after 5 minutes
        if hasattr(socket, "TCP_KEEPIDLE"):
            self.socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 300)

    def addSuites(self, *args):
        # use this as an opportunity to broadcast our address
        serverAddress = self.getAddress()
        self.diag.info("Starting slave server at " + serverAddress)
        # Tell the submission code where we are
        QueueSystemServer.instance.setSlaveServerAddress(serverAddress)

    def handlerClass(self):
        return SlaveRequestHandler

    def canBeMainThread(self):
        return False  # We wait for sockets and stuff

    def run(self):
        while not self.terminate:
            self.diag.info("Waiting for a new request...")
            try:
                self.handle_request()
            except:
                # e.g. can get interrupted system call here in 'select' if we get a signal
                sys.stderr.write("WARNING: slave server caught exception while processing request!\n")
                plugins.printException()

        self.diag.info("Terminating slave server")

    def notifyAllRead(self, *args):
        if len(self.testMap) == 0:
            self.notifyAllComplete()

    def notifyAllComplete(self):
        self.diag.info("Notified all complete, shutting down soon...")
        self.terminate = True
        sendSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sendSocket.connect(self.socket.getsockname())
        sendSocket.sendall("TERMINATE_SERVER\n".encode(getpreferredencoding()))
        sendSocket.close()

    def getAddress(self):
        host, port = self.socket.getsockname()
        return host + ":" + str(port)

    def notifyAdd(self, test, initial):
        if test.classId() == "test-case":
            self.storeTest(test)

    def storeTest(self, test):
        testPath = test.getRelPath()
        testApp = test.app.name + test.app.versionSuffix()
        if testApp not in self.testMap:
            self.testMap[testApp] = {}
        self.testMap[testApp][testPath] = test
        self.testLocks[test] = RLock()

    def changeState(self, test, state):
        # Several threads could be trying to do this at once...
        lock = self.testLocks.get(test)
        with lock:
            allow = self.allowChange(test.state, state)
            if allow:
                test.changeState(state)
            else:
                self.diag.info("Rejecting state change, old state " + test.state.category +
                               " is complete, new state " + state.category + " is not.")
        return allow

    def changeStateOrRerun(self, test, state, rerun):
        oldBt = test.state.briefText
        self.diag.info("Changed from '" + oldBt + "' to '" + state.briefText + "'")
        if rerun and self.totalReruns < self.maxReruns:
            lock = self.testLocks.get(test)
            with lock:
                self.totalReruns += 1
            self.diag.info("Instructed to rerun test " + test.uniqueName + ", now performed " +
                           str(self.totalReruns) + " reruns of max " + str(self.maxReruns) + ".")
            QueueSystemServer.instance.queueTestForRerun(test)
            return True
        else:
            if rerun:
                self.diag.info("Instructed to rerun test " + test.uniqueName +
                               ", but refusing, already rerun maximum " + str(self.totalReruns) + " times.")
            self.changeState(test, state)
            return False

    def allowChange(self, oldState, newState):
        return newState.isComplete() or not oldState.isComplete()

    def getTest(self, testString):
        self.diag.info("Received request for '" + testString + "'")
        try:
            appName, testPath = socketParse(testString)
            return self.testMap[appName][testPath]
        except ValueError:
            return

    def getFilePushProcess(self, test, userAndHost, path):
        key = userAndHost, path
        with self.filePushLock:
            if key in self.filePushProcesses:
                return self.filePushProcesses[key], False
            else:
                proc = test.app.getRemoteCopyFileProcess(path, "localhost", os.path.dirname(path), userAndHost)
                self.filePushProcesses[key] = proc
                return proc, True

    def pushFiles(self, test, userAndHost, paths):
        for path in paths:
            proc, started = self.getFilePushProcess(test, userAndHost, path)
            if started:
                self.diag.info("Pushing '" + path + "'...")
                proc.wait()
                self.diag.info("Done Pushing '" + path + "'")
                # Aim for synchronising tests properly
                QueueSystemServer.instance.sendServerState("Sychronised " + path + " to " + userAndHost)
            else:
                self.diag.info("Waiting for '" + path + "'...")
                while proc.poll() is None:
                    time.sleep(0.1)
                self.diag.info("Done Waiting for '" + path + "'.")


class MasterTextResponder(TextDisplayResponder):
    def getPrefix(self, test):
        return "S: "  # don't get things in order, so indenting is pointless

    def getTestRepr(self, test):
        return test.relPathRepr() # don't get things in order, can't rely on that to show where in the structure we are

    def shouldDescribe(self, test):
        return True  # Write the successful tests also


class MasterInteractiveResponder(InteractiveResponder):
    def getPrefix(self, test):
        return ""  # don't get things in order, so indenting is pointless




import os
import importlib
from texttestlib import plugins


def getVersionControlConfig(apps, inputOptions):
    allDirs = [app.getDirectory() for app in apps] + inputOptions.rootDirectories
    for dir in allDirs:
        # Hack for self-testing...
        dirToCheck = dir if os.path.basename(dir) in os.getenv(
            "TEXTTEST_SELFTEST_DIR_NAMES", "").split(",") else os.path.realpath(dir)
        config = getConfigFromDirectory(dirToCheck)
        if config:
            return config


def getConfigFromDirectory(directory):
    allEntries = os.listdir(directory)
    for controlDirName in plugins.controlDirNames:
        if controlDirName in allEntries:
            module = controlDirName.lower().replace(".", "")
            try:
                controlDir = os.path.join(directory, controlDirName)
                if module != "cvs" or not os.path.isdir(os.path.join(controlDir, "CVS")):
                    # Avoid overarching directories "CVS" which are not control directories...
                    mod = importlib.import_module("." + module, __name__)
                    return mod.InteractiveActionConfig(controlDir)
            except ImportError:  # There may well be more VCSs than we have support for...
                pass
    dirname = os.path.dirname(directory)
    if dirname != directory:
        return getConfigFromDirectory(dirname)
